import numpy as np

from tqdm import tqdm
from sklearn.utils import check_array
from sklearn.utils.validation import check_is_fitted
from sklearn.preprocessing import normalize
from pyod.models.base import BaseDetector
from scipy.special import softmax
from uni_norm_dist_omp import uni_norm_dist_omp, compute_proba, compute_obj


class GaussDL(BaseDetector):
    def __init__(self,
                 contamination,
                 n_components=100,
                 n_nonzero_coefs=3,
                 max_iter=50,
                 D0=None,
                 sigmas=None,
                 _lambda=None,
                 decision_function_name=None,
                 random_state=None):
        self.D = None
        self.decision_scores_ = None

        self.contamination = contamination
        self.n_components = n_components
        self.n_nonzero_coefs = n_nonzero_coefs
        self.max_iter = max_iter
        self.D0 = D0
        self.sigmas = sigmas
        self._lambda = _lambda
        self.decision_function_name = decision_function_name
        self.random_state = random_state
        
        self.errs = np.array([])
        self.probas = np.array([])
        self.objs = np.array([])

    def fit(self, X, y=None):
        """Fit detector. y is ignored in unsupervised methods.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples.
        y : Ignored
            Not used, present for API consistency by convention.
        Returns
        -------
        self : object
            Fitted estimator.
        """
        # validate inputs X and y (optional)
        Y = check_array(X.T)
        self._set_n_classes(y)
        
        n_components = self.n_components
        n_features, n_samples = Y.shape
        error_ = np.zeros((1, n_samples))
        
        repr_err_ = np.zeros((n_samples, self.max_iter))
        proba_ = np.zeros((n_samples, self.max_iter))
        obj_ = np.zeros((n_samples, self.max_iter))

        if self.D0 is None:
            self.D = normalize(self.random_state.randn(Y.shape[0], self.n_components), axis=0, norm='l2')
        else:
            self.D = self.D0

        for i_iter in tqdm(range(self.max_iter)):
            D_new = np.zeros((n_features, n_components))
            atoms_counter = np.zeros((1, n_components))
            
            for i_sample in range(n_samples):
                # extract the current sample
                y = Y[:, i_sample:i_sample+1]
            
                # compute univariate normal distribution
                X, alphas, support, errs = uni_norm_dist_omp(y, self.D, self.sigmas, self.n_nonzero_coefs, _lambda=self._lambda, max_iter=1)
                error_[0, i_sample] = errs[-1]
                
                # update D_new atoms
                D_new[:, support] += X
                atoms_counter[:, support] += 1

                # update current residual
                repr_err_[i_sample, i_iter] += errs[-1]
                proba_[i_sample, i_iter] += compute_proba(X, self.D, self.sigmas, support)
                obj_[i_sample, i_iter] += compute_obj(X, self.D, self.sigmas, support, y, alphas, self._lambda)

            # compute the new dictionary
            idx = np.where(atoms_counter[0, :] != 0)[0]
            D_new[:, idx] /= atoms_counter[0, idx]
            idx = np.where(atoms_counter[0, :] == 0)[0]
            D_new[:, idx] = self.D[:, idx]
            self.D = normalize(D_new, axis=0, norm='l2')

        self.repr_err_ = repr_err_
        self.proba_ = proba_
        self.obj_ = obj_
        
        self.decision_scores_ = error_
        self._process_decision_scores()
        return self

    def decision_function(self, X):
        """Predict raw anomaly score of X using the fitted detector.
        The anomaly score of an input sample is computed based on different
        detector algorithms. For consistency, outliers are assigned with
        larger anomaly scores.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only
            if they are supported by the base estimator.
        Returns
        -------
        anomaly_scores : numpy array of shape (n_samples,)
            The anomaly score of the input samples.
        """
        probas = np.array([])
        Y = check_array(X.T)
        check_is_fitted(self, ['decision_scores_', 'threshold_', 'labels_'])
        
        if self.errs.size == 0 and self.probas.size == 0 and self.objs.size == 0:
            sigmas = self.sigmas
            
            for i in tqdm(range(Y.shape[1])):
                y = Y[:, i:i+1]
                # compute univariate normal distribution
                X, alphas, support, errs = uni_norm_dist_omp(y, self.D, sigmas, self.n_nonzero_coefs, self._lambda)
                
                err = errs[-1]
                proba = compute_proba(X, self.D, sigmas, support)
                obj = compute_obj(X, self.D, sigmas, support, y, alphas, self._lambda)
                
                self.errs = np.append(self.errs, err)
                self.probas = np.append(self.probas, proba)
                self.objs = np.append(self.objs, obj)

        if self.decision_function_name == 'err':
            probas = self.errs
        elif self.decision_function_name == 'proba':
            probas = self.probas
        else:  # 'obj_value'
            probas = self.objs

        return probas
