# Author: Script by Yue Zhao <zhaoy@cmu.edu>, adapted by Denis C. Ilie-Ablachim <denis.ilie_ablachim@upb.ro>, Andra Baltoiu <andra.baltoiu@upb.ro> and Bogdan Dumitrescu <bogdan.dumitrescu@upb.ro> for "Sparse representation with Gaussian atoms"
# License: BSD 2 clause

from __future__ import division
from __future__ import print_function

import os
import sys
from time import time

# temporary solution for relative imports in case pyod is not installed
# if pyod is installed, no need to use the following line
sys.path.append(
    os.path.abspath(os.path.join(os.path.dirname("__file__"), '..')))
# supress warnings for clean output
import warnings

warnings.filterwarnings("ignore")

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from scipy.io import loadmat

from GaussDL import GaussDL
from GaussOMP import GaussOMP

from pyod.utils.utility import standardizer
from pyod.utils.utility import precision_n_scores

from sklearn.metrics import roc_auc_score
from sklearn.preprocessing import normalize

n_components_features_ratio_list = [2]
n_nonzero_coefs_list = [2]
max_iter_list = [40]
lambda_list = [100]
sigmas_list = [5e-2]
# sigmas_min_list = [0.05]
# sigmas_max_list = [0.1]

fit_algorithm = 'aksvd'
transform_algorithm = 'omp'

# Define data file and read X and y
mat_file_list = [
    "dependency_annthyroid.mat",
    # "dependency_breastw.mat",
    # "dependency_cardio.mat",
    # "dependency_cover.mat",
    # "dependency_fault.mat",
    # "dependency_glass.mat",
    # "dependency_Hepatitis.mat",
    # "dependency_Ionosphere.mat",
    # "dependency_landsat.mat",
    # "dependency_letter.mat",
    # "dependency_Lymphography.mat",
    # "dependency_magic.gamma.mat",
    # "dependency_mammography.mat",
    # "dependency_musk.mat",
    # "dependency_pendigits.mat",
    # "dependency_Pima.mat",
    # "dependency_satellite.mat",
    # "dependency_satimage.mat",
    # "dependency_SpamBase.mat",
    # "dependency_Stamps.mat",
    # "dependency_thyroid.mat",
    # "dependency_vertebral.mat",
    # "dependency_vowels.mat",
    # "dependency_Waveform.mat",
    # "dependency_WBC.mat",
    # "dependency_WDBC.mat",
    # "dependency_Wilt.mat",
    # "dependency_wine.mat",
    # "dependency_WPBC.mat",
    # "dependency_yeast.mat"
]

# define the number of iterations
n_ite = 1
n_classifiers = 6

df_columns = [
    'Data', '#Samples', '# Dimensions', 'Outlier Perc',
    'GaussOMP[err]', 'GaussOMP[proba]', 'GaussOMP[obj]',
    'GaussDL[err]', 'GaussDL[proba]', 'GaussDL[obj]',
]

clf_name_map = {
    'Gauss Orthogonal Matching Pursuit (GaussOMP)': 'GaussOMP',
    'Gauss Dictionary Learning (GaussDL)': 'GaussDL',
}


for n_components_features_ratio in n_components_features_ratio_list:    
    for n_nonzero_coefs in n_nonzero_coefs_list:
        for max_iter in max_iter_list:
            for _lambda in lambda_list:
                for sigma in sigmas_list:
                    ext_name = '_n_ratio_{}_s__prec_{}_max_iter_{}_lambda_{}_sigma_{}.csv'.format(n_components_features_ratio, n_nonzero_coefs, max_iter, _lambda, sigma)
                    if os.path.exists('./results/roc' + ext_name):
                        continue

                    # initialize the container for saving the results
                    roc_df = pd.DataFrame(columns=df_columns)
                    prn_df = pd.DataFrame(columns=df_columns)
                    time_df = pd.DataFrame(columns=df_columns)
                    
                    for j in range(len(mat_file_list)):
                    
                        mat_file = mat_file_list[j]
                        mat = loadmat(os.path.join('dependency_mats', mat_file))

                        X = mat['Y']
                        y = mat['labels'].ravel()
                        
                        outliers_fraction = np.count_nonzero(y) / len(y)
                        outliers_percentage = round(outliers_fraction * 100, ndigits=4)
                    
                        # compute the number of atoms
                        n_features = X.shape[1]
                        n_components = int(n_features * n_components_features_ratio)
                        
                        # compute sigmas
                        sigmas = np.ones(n_components) * sigma
                        # sigmas = sigma_min + random_state.rand(n_components)*(sigma_max - sigma_min)
                    
                        # construct containers for saving results
                        roc_list = [mat_file[:-4], X.shape[0], X.shape[1], outliers_percentage]
                        prn_list = [mat_file[:-4], X.shape[0], X.shape[1], outliers_percentage]
                        time_list = [mat_file[:-4], X.shape[0], X.shape[1], outliers_percentage]
                    
                        roc_mat = np.zeros([n_ite, n_classifiers])
                        prn_mat = np.zeros([n_ite, n_classifiers])
                        time_mat = np.zeros([n_ite, n_classifiers])
                    
                        for i in range(n_ite):
                            print("\n... Processing", mat_file, '...', 'Iteration', i + 1)
                            random_state = np.random.RandomState(i)
                    
                            # initialize starting dictionary
                            D0 = normalize(np.random.randn(n_features, n_components), axis=0, norm='l2')
                    
                            # standardizing data for processing
                            X_norm = standardizer(X)

                            # 70% data for training and 30% for testing
                            X_train_norm, X_test_norm, y_train, y_test = \
                                train_test_split(X_norm, y, test_size=0.3, random_state=random_state)
                    
                            classifiers = {
                                'Gauss Orthogonal Matching Pursuit (GaussOMP)': GaussOMP(contamination=outliers_fraction,
                                    n_components=n_components, n_nonzero_coefs=n_nonzero_coefs,
                                    fit_algorithm=fit_algorithm, transform_algorithm=transform_algorithm,
                                    max_iter=max_iter, D0=D0, sigmas=sigmas, _lambda=_lambda,
                                    decision_function_name='err', random_state=random_state),
                                'Gauss Dictionary Learning (GaussDL)': GaussDL(contamination=outliers_fraction,
                                    n_components=n_components, n_nonzero_coefs=n_nonzero_coefs,
                                    max_iter=max_iter, D0=D0, sigmas=sigmas, _lambda=_lambda,
                                    decision_function_name='err', random_state=random_state),
                            }
    
                            classifiers_indices = {
                                'Gauss Orthogonal Matching Pursuit (GaussOMP)': 0,
                                'Gauss Dictionary Learning (GaussDL)': 3,
                            }
                    
                            for clf_name, clf in classifiers.items():
                                if clf_name in ['Gauss Orthogonal Matching Pursuit (GaussOMP)',
                                                'Gauss Dictionary Learning (GaussDL)']:
                                    t0 = time()
                                    clf.fit(X_train_norm)
                                    t1 = time()
                                    train_duration = round(t1 - t0, ndigits=4)

                                    # Gauss Dictionary Learning (GaussDL[err])
                                    clf.decision_function_name = 'err'
                                    t0 = time()
                                    test_scores = clf.decision_function(X_test_norm)
                                    t1 = time()
                                    duration = round(t1 - t0, ndigits=4) + train_duration
                                    
                                    roc = round(roc_auc_score(y_test, test_scores), ndigits=4)
                                    prn = round(precision_n_scores(y_test, test_scores), ndigits=4)
                                    
                                    print('{clf_name} ROC:{roc}, precision @ rank n:{prn}, '
                                          'execution time: {duration}s'.format(
                                        clf_name=clf_name_map[clf_name] + '[err]',
                                        roc=roc, prn=prn, duration=duration))
                                    
                                    time_mat[i, classifiers_indices[clf_name]] = duration
                                    roc_mat[i, classifiers_indices[clf_name]] = roc
                                    prn_mat[i, classifiers_indices[clf_name]] = prn

                                    # Gauss Dictionary Learning (GaussDL[proba]) 
                                    clf.decision_function_name = 'proba'
                                    t0 = time()
                                    test_scores = clf.decision_function(X_test_norm)
                                    t1 = time()
                                    duration = round(t1 - t0, ndigits=4) + train_duration
    
                                    roc = round(roc_auc_score(y_test, test_scores), ndigits=4)
                                    prn = round(precision_n_scores(y_test, test_scores), ndigits=4)
                                    
                                    print('{clf_name} ROC:{roc}, precision @ rank n:{prn}, '
                                          'execution time: {duration}s'.format(
                                        clf_name=clf_name_map[clf_name] + '[proba]',
                                        roc=roc, prn=prn, duration=duration))
                                    
                                    time_mat[i, classifiers_indices[clf_name] + 1] = duration
                                    roc_mat[i, classifiers_indices[clf_name] + 1] = roc
                                    prn_mat[i, classifiers_indices[clf_name] + 1] = prn
    
                                    # Gauss Dictionary Learning (GaussDL[obj]) 
                                    clf.decision_function_name = 'obj'
                                    t0 = time()
                                    test_scores = clf.decision_function(X_test_norm)
                                    t1 = time()
                                    duration = round(t1 - t0, ndigits=4) + train_duration
    
                                    roc = round(roc_auc_score(y_test, test_scores), ndigits=4)
                                    prn = round(precision_n_scores(y_test, test_scores), ndigits=4)
                                    
                                    print('{clf_name} ROC:{roc}, precision @ rank n:{prn}, '
                                          'execution time: {duration}s'.format(
                                        clf_name=clf_name_map[clf_name] + '[obj]',
                                        roc=roc, prn=prn, duration=duration))
                                    
                                    time_mat[i, classifiers_indices[clf_name] + 2] = duration
                                    roc_mat[i, classifiers_indices[clf_name] + 2] = roc
                                    prn_mat[i, classifiers_indices[clf_name] + 2] = prn
                                else:
                                    t0 = time()
                                    clf.fit(X_train_norm)
                                    test_scores = clf.decision_function(X_test_norm)
                                    t1 = time()
                                    duration = round(t1 - t0, ndigits=4)
                        
                                    roc = round(roc_auc_score(y_test, test_scores), ndigits=4)
                                    prn = round(precision_n_scores(y_test, test_scores), ndigits=4)
                        
                                    print('{clf_name} ROC:{roc}, precision @ rank n:{prn}, '
                                          'execution time: {duration}s'.format(
                                        clf_name=clf_name, roc=roc, prn=prn, duration=duration))
                        
                                    time_mat[i, classifiers_indices[clf_name]] = duration
                                    roc_mat[i, classifiers_indices[clf_name]] = roc
                                    prn_mat[i, classifiers_indices[clf_name]] = prn
                    
                        time_list = time_list + np.mean(time_mat, axis=0).tolist()
                        temp_df = pd.DataFrame(time_list).transpose()
                        temp_df.columns = df_columns
                        time_df = pd.concat([time_df, temp_df], axis=0)

                        roc_list = roc_list + np.mean(roc_mat, axis=0).tolist()
                        temp_df = pd.DataFrame(roc_list).transpose()
                        temp_df.columns = df_columns
                        roc_df = pd.concat([roc_df, temp_df], axis=0)

                        prn_list = prn_list + np.mean(prn_mat, axis=0).tolist()
                        temp_df = pd.DataFrame(prn_list).transpose()
                        temp_df.columns = df_columns
                        prn_df = pd.concat([prn_df, temp_df], axis=0)

                        # Save the results for each run
                        ext_name = '_n_ratio_{}_s__prec_{}_max_iter_{}_lambda_{}_sigma_{}.csv'.format(n_components_features_ratio, n_nonzero_coefs, max_iter, _lambda, sigma)
                        time_df.to_csv('./results/time' + ext_name, index=False, float_format='%.3f')
                        roc_df.to_csv('./results/roc' + ext_name, index=False, float_format='%.3f')
                        prn_df.to_csv('./results/prc' + ext_name, index=False, float_format='%.3f')
