import math
import functools
import numpy as np
import matplotlib.pyplot as plt

from sklearn.preprocessing import normalize


def compute_proba(X, D, sigmas, support):
    n_features = D.shape[0]
    probs = np.linalg.norm(D[:, support] - X, axis=0)**2 / (2*sigmas[support]**2) + n_features*np.log(sigmas[support] * np.sqrt(2*np.pi))
    return sum(probs)


def compute_repr_err(y, X, alphas):
    err = np.linalg.norm(y - np.dot(X, alphas))
    return err


def compute_obj(X, D, sigmas, support, y, alphas, _lambda):
    obj_value = np.sum(np.linalg.norm(D[:, support] - X, axis=0)**2 / (sigmas[support]**2))
    obj_value += _lambda * np.linalg.norm(y - np.dot(X, alphas))
    return obj_value


def obj_func_template(x, d, y_tilde, sigma, _lambda):
    alpha = np.dot(x.T, y_tilde) / np.linalg.norm(x)**2
    return np.linalg.norm(x - d)**2 / sigma**2 + _lambda * np.linalg.norm(y_tilde - alpha * x)**2


def obj_grad_template(x, d, y_tilde, sigma, _lambda):
    alpha = np.dot(x.T, y_tilde) / np.linalg.norm(x)**2
    return (2/sigma**2 * (x - d) - 2*_lambda*alpha * (y_tilde - alpha * x)).T


def gamma_obj_func_template(gamma, d, y_tilde, obj_func):
    x = gamma * d + (1 - gamma) * y_tilde * np.sign(np.dot(d.T, y_tilde))
    x = x / np.linalg.norm(x)
    return obj_func(x)


def bisection_method_grad(obj_grad, gamma_left, gamma_right, d, y_tilde, eps):
    # compute optimization direction
    sign = np.sign(np.dot(d.T, y_tilde))
    y_tilde = sign * y_tilde
    direction = y_tilde - d
    
    # compute the two sides of the bisection
    x_left = gamma_left * d + (1 - gamma_left) * y_tilde
    x_right = gamma_right * d + (1 - gamma_right) * y_tilde
    
    # normalize vectors
    x_left = x_left / np.linalg.norm(x_left)
    x_right = x_right / np.linalg.norm(x_right)
    
    # compute objective gradients
    df_x_left = np.dot(obj_grad(x_left), direction)
    df_x_right = np.dot(obj_grad(x_right), direction)
    
    while abs(gamma_left - gamma_right) > eps:
        gamma_mid = (gamma_left + gamma_right) / 2 
        x_mid = gamma_mid * d + (1 - gamma_mid) * y_tilde
        x_mid = x_mid / np.linalg.norm(x_mid)
        df_x_mid = np.dot(obj_grad(x_mid), direction)
        
        if np.sign(df_x_left * df_x_mid) < 0:
            gamma_right = gamma_mid
            df_x_right = df_x_mid
        elif np.sign(df_x_right * df_x_mid) < 0:
            gamma_left = gamma_mid
            df_x_left = df_x_mid

    gamma_mid = (gamma_left + gamma_right) / 2
    return gamma_mid


def gss(f, a, b, tol=1e-5):
    """Golden-section search.

    Given a function f with a single local minimum in
    the interval [a,b], gss returns a subset interval
    [c,d] that contains the minimum with d-c <= tol.

    Example:
    >>> f = lambda x: (x-2)**2
    >>> a = 1
    >>> b = 5
    >>> tol = 1e-5
    >>> (c,d) = gss(f, a, b, tol)
    >>> print(c, d)
    1.9999959837979107 2.0000050911830893
    """

    invphi = (math.sqrt(5) - 1) / 2  # 1 / phi
    invphi2 = (3 - math.sqrt(5)) / 2  # 1 / phi^2

    (a, b) = (min(a, b), max(a, b))
    h = b - a
    if h <= tol:
        return (a, b)

    # Required steps to achieve tolerance
    n = int(math.ceil(math.log(tol / h) / math.log(invphi)))

    c = a + invphi2 * h
    d = a + invphi * h
    yc = f(c)
    yd = f(d)

    for k in range(n - 1):
        if yc < yd:  # yc > yd to find the maximum
            b = d
            d = c
            yd = yc
            h = invphi * h
            c = a + invphi2 * h
            yc = f(c)
        else:
            a = c
            c = d
            yc = yd
            h = invphi * h
            d = a + invphi * h
            yd = f(d)

    if yc < yd:
        return (a, d)
    else:
        return (c, b)


def compute_nearest_uni_norm_dist_atom(y_tilde, d, sigma, _lambda, tol=1e-5):
    # bisection method interval
    gamma_left = 0
    gamma_right = 1 - gamma_left

    # custom bisection method
    # obj_grad = functools.partial(obj_grad_template, d=d, y_tilde=y_tilde, sigma=sigma, _lambda=_lambda)
    # gamma = bisection_method_grad(obj_grad, gamma_left, gamma_right, d, y_tilde, tol)
    
    # custom golden section method
    obj_func = functools.partial(obj_func_template, d=d, y_tilde=y_tilde, sigma=sigma, _lambda=_lambda)
    gamma_func = functools.partial(gamma_obj_func_template, d=d, y_tilde=y_tilde, obj_func=obj_func)
    gamma_left, gamma_right = gss(gamma_func, gamma_left, gamma_right, tol)
    gamma = (gamma_left + gamma_right) / 2

    # update current x
    x = gamma * d + (1 - gamma) * y_tilde * np.sign(np.dot(d.T, y_tilde))
    x = x / np.linalg.norm(x)

    # update alpha coef
    alpha = np.dot(x.T, y_tilde) / np.linalg.norm(x)**2
    
    return x, alpha


def _uni_norm_dist_omp(y, D, sigmas, s, _lambda=1e-3, max_iter=10):
    # method of global univariate normal distribution OMP
    
    # init Xs solutions
    n_features = D.shape[0]
    X = np.random.randn(n_features, s)
    X = normalize(X, axis=0, norm='l2')
    
    # initial alpha solution for the LSM
    alphas, _, _, _ = np.linalg.lstsq(X, y, rcond=None)

    # compute current representation error
    errs = np.array([])
    error = y - np.dot(X, alphas)
    residual = y / np.linalg.norm(y)

    # compute residual distance from each Gaussian distribution
    atoms_direction = np.sign(np.dot(residual.T, D))
    distances = np.linalg.norm(atoms_direction * D - residual, axis=0)**2 / (2*sigmas**2) + n_features*np.log(sigmas)

    # define sparse learning support
    support = np.argsort(distances)[:s]
    
    for _ in range(max_iter):
        for index, atom_index in enumerate(support):
            # Compute the most representative atom from the Gaussian distribution
            x = X[:, index:index + 1]
            d = D[:, atom_index:atom_index + 1]
            alpha = alphas[index]
            sigma = sigmas[atom_index]

            # modify error
            y_tilde = error + alpha * x

            # compute the nearest gaussian atom
            x, alpha = compute_nearest_uni_norm_dist_atom(y_tilde, d, sigma, _lambda)

            X[:, index:index + 1] = x
            alphas[index] = alpha

            # update representation error
            error = y_tilde - alpha * x

        errs = np.append(errs, np.sum(np.linalg.norm(X - D[:, support], axis=0)**2 / sigmas[support]**2) +
                                      _lambda * np.linalg.norm(y - np.dot(X, alphas))**2)
    
    return X, alphas, support, errs


def uni_norm_dist_omp(y, D, sigmas, s, _lambda=1e-1, _lambda_rate=1.01, max_iter=10, update_lambda=False):
    # define sparse learning support
    support = []
    errs = np.array([])
    error = y
    
    # init Xs solutions
    n_features = D.shape[0]
    X = np.random.randn(n_features, s)
    X = normalize(X, axis=0, norm='l2')
    
    # init alphas
    alphas = np.zeros((s, 1))
    
    for support_index in range(s):
        # compute residual distance from each Gaussian distribution
        norm_error = error #/ np.linalg.norm(error)
        atoms_direction = np.sign(np.dot(norm_error.T, D))
        distances = np.linalg.norm(atoms_direction * D - norm_error, axis=0)**2 / (2*sigmas**2) + n_features*np.log(sigmas)

        # avoid atom duplication
        distances[support] = np.Inf

        # find the nearest Gaussian atoms
        j_atom = np.argmin(distances)
        support.append(j_atom)

        # initial alpha solution for the LSM
        error = error - X[:, support_index:support_index+1] * alphas[support_index]

        for _ in range(max_iter):
            for index, atom_index in enumerate(support):
                # Compute the most representative atom from the Gaussian distribution
                x = X[:, index:index + 1]
                d = D[:, atom_index:atom_index + 1]
                alpha = alphas[index]
                sigma = sigmas[atom_index]
    
                # modify error
                y_tilde = error + alpha * x
    
                # compute the nearest gaussian atom
                x, alpha = compute_nearest_uni_norm_dist_atom(y_tilde, d, sigma, _lambda)
    
                X[:, index:index + 1] = x
                alphas[index] = alpha
    
                # update representation error
                error = y_tilde - alpha * x
    
            # update lambda value
            if update_lambda:
                _lambda *= _lambda_rate
            
            error = y - np.dot(X[:, :support_index + 1], alphas[:support_index + 1, :])
            errs = np.append(errs, np.linalg.norm(error))

    return X, alphas, support, errs


def main():
    s = 3
    n_atoms = 500
    n_features = 64
    sigma = 0.1
    
    x_thresh = 1
    noise_v = 1e-3
    
    # init dictionary atoms
    D = np.random.randn(n_features, n_atoms)
    D = normalize(D, axis=0, norm='l2')
    
    # init atoms variances
    # sigmas = np.random.rand(n_atoms)
    sigmas = np.ones((n_atoms)) * sigma
    
    # init constraint solution y
    support_origin = list(range(s))
    support_origin = support_origin[:s]
    coef = np.random.randn(s, 1)
    coef = coef + np.sign(coef)*x_thresh
    y = np.dot(D[:, support_origin], coef) + noise_v * np.random.randn(n_features, 1)
    
    # compute univariate normal distribution
    X, alphas, support, errs = uni_norm_dist_omp(y, D, sigmas, s)
    
    plt.figure()
    plt.plot(range(len(errs)), errs)
    plt.show()
    plt.xlabel('iters')
    plt.ylabel('error')
    plt.title('OMP - stage update')


if __name__ == '__main__':
    main()
